'use strict';

var AJAX_TIMEOUT = 60000;

module.exports = ajax;

function ajax(options, next) {
  var url = options.url,
      proxy = options.proxy;

  var XHR = options.xhr || XMLHttpRequest;
  var method = (options.method || 'GET').toUpperCase();
  // eslint-disable-next-line new-cap
  var xhr = new XHR();
  var async = options.sync !== true;
  var timeout = +options.timeout || AJAX_TIMEOUT;
  var ajaxTimeout = void 0;

  /**
   * Wraps callback functions to remove xhr data.
   *
   * @param  {Function} fn
   * @return {Function}
   */
  function complete(fn) {
    return function () {
      window.clearTimeout(ajaxTimeout);

      // Remove references to unused functions.
      xhr.onload = null;
      xhr.onerror = null;
      xhr.onabort = null;

      for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      return fn.apply(this, args);
    };
  }

  xhr.open(method, '' + (proxy || '') + url, async);

  // Sets all request headers before we make the request.
  Object.keys(options.headers).forEach(function (key) {
    xhr.setRequestHeader(key, options.headers[key]);
  });

  // Enable hooking into the ajax request before we send it.
  if (options.beforeSend) {
    options.beforeSend(xhr);
  }

  // Successful callback.
  xhr.onload = complete(function () {
    return next(null, buildResponse(xhr));
  });

  // Failure callback.
  xhr.onabort = complete(function () {
    return next(new Error(xhr.statusText || 'AJAX request aborted'), buildResponse(xhr));
  });

  xhr.onerror = complete(function () {
    return next(new Error(xhr.statusText || 'AJAX request errored'), buildResponse(xhr));
  });

  // Set a request timeout. Modern browsers can set a `timeout` property
  // which works the same, but we'll use a timeout for consistency.
  if (async) {
    ajaxTimeout = window.setTimeout(complete(function () {
      xhr.abort();

      // Calls the `next` function with the timeout details.
      return next(new Error('AJAX timeout of ' + timeout + 'ms exceeded'), buildResponse(xhr));
    }), timeout);
  }

  xhr.send(options.body);
}

function buildResponse(xhr) {
  var headers = {};

  xhr.getAllResponseHeaders().trim().split('\n').forEach(function (header) {
    var index = header.indexOf(':');
    var key = header.slice(0, index).trim().toLowerCase();
    var value = header.slice(index + 1).trim();

    headers[key] = value;
  });

  return {
    body: xhr.responseText,
    statusCode: xhr.status,
    method: xhr.method,
    rawRequest: xhr,
    headers: headers
  };
}
//# sourceMappingURL=ajax.js.map