'use strict';

var template = require('./generator/template');
var triggerAuthentication = require('./authentication');

/**
 * Map authentication types to automatic preference.
 *
 * @type {Array}
 */
var ORDER_PREFERENCE = ['OAuth 2.0', 'OAuth 1.0', 'Basic Authentication'];

/**
 * Required authentication keys used to check the options object.
 *
 * @type {Object}
 */
var DEFAULT_REQUIRED_FIELDS = {
  'OAuth 1.0': {
    consumerKey: true,
    consumerSecret: true
  },
  'OAuth 2.0': {
    clientId: true,
    clientSecret: true
  },
  'Basic Authentication': {
    username: true,
    password: true
  }
};

/**
 * Possible tokens to be filled out by the user.
 *
 * @type {Object}
 */
var DEFAULT_POSSIBLE_FIELDS = {
  'OAuth 1.0': ['consumerKey', 'consumerSecret'],
  'OAuth 2.0': ['clientId', 'clientSecret', 'scopes'],
  'Basic Authentication': ['username', 'password']
};

/**
 * These fields must be hidden in the UI.
 *
 * @type {Object}
 */
var SECRET_FIELDS = {
  clientSecret: true,
  consumerSecret: true,
  password: true
};

/**
 * Returns an object of available keys and whether they are required.
 *
 * @param  {Object} scheme
 * @return {Object}
 */
var requiredTokens = function requiredTokens(scheme) {
  var keys = Object.assign({}, DEFAULT_REQUIRED_FIELDS[scheme.type]);

  // Special case is required for OAuth2 implicit auth flow.
  if (scheme.type === 'OAuth 2.0') {
    keys.clientSecret = !scheme.settings.authorizationGrants.some(function (grant) {
      return grant === 'token' || grant === 'implicit';
    });
  }

  return keys;
};

/**
 * Sanitize scopes to be an array.
 *
 * @param  {*}     scopes
 * @return {Array}
 */
var sanitizeScope = function sanitizeScope(scopes) {
  if (typeof scopes === 'string') {
    return scopes.split(' ');
  }

  if (!Array.isArray(scopes)) {
    return [];
  }

  return scopes;
};

/**
 * Prompt the user for authentication tokens based on a scheme. We can pass in
 * the current options object to help decide what data to display to the user.
 *
 * @param {Object}   scheme
 * @param {Object}   options
 * @param {Function} done
 */
var promptTokens = function promptTokens(scheme, options) {
  var needsTokens = requiredTokens(scheme);

  // Generate an array of the tokens to use with our prompt and filter
  // explicitly not required tokens.
  var possibleTokens = DEFAULT_POSSIBLE_FIELDS[scheme.type].filter(function (token) {
    return needsTokens[token] !== false;
  });

  // Multiple ways of setting the scope option.
  options.scopes = sanitizeScope(options.scope || options.scopes);
  delete options.scope;

  var prompts = {};
  prompts.keys = [];

  possibleTokens.forEach(function (key) {
    if (key === 'scopes') {
      var scopes = sanitizeScope(scheme.settings.scopes);

      if (scopes && scopes.length) {
        prompts.scopes = scopes;
        prompts.keys.push(key);
      }
    } else {
      // DEFAULT_FIELD_NAMES[key];
      prompts.keys.push(key);
    }
  });

  if (prompts.keys.every(function (key) {
    return options[key] || options.credentials && options.credentials[key];
  })) {
    return Promise.resolve(Object.assign({}, options.credentials, options));
  }

  if (options.prompt) {
    return options.prompt(prompts).then(function ($prompts) {
      return Object.assign({}, options.credentials, $prompts, options);
    });
  }

  return Promise.reject(new Error('Missing required fields: ' + prompts.keys.filter(function (key) {
    return !options[key];
  }).join(', ')));
};

/**
 * Authenticate using an authentication scheme and passed in options.
 *
 * @param  {Object}   scheme
 * @param  {Object}   options
 * @param  {Function} done
 */
var authenticate = function authenticate(scheme, options) {
  var opts = Object.assign({
    type: scheme.type
  }, options, scheme.settings);

  // Interpolate all uri parameters.
  ['accessTokenUri', 'authorizationUri', 'requestTokenUri', 'authorizationUri', 'tokenCredentialsUri'].forEach(function (param) {
    opts[param] = template(opts[param], opts.baseUriParameters);
  });

  return triggerAuthentication(opts).then(function (tokens) {
    return { scheme: scheme, options: options, tokens: tokens };
  });
};

/**
 * Return the preferred scheme option from an object of every scheme.
 *
 * @param  {Object} schemes
 * @return {Object}
 */
var preferredScheme = function preferredScheme(schemes) {
  // Fix `undefined` schemes issues.
  if (!schemes || !Object.keys(schemes).length) {
    return null;
  }

  var method = ORDER_PREFERENCE.find(function (preference) {
    var intersectable = Object.keys(schemes).map(function (scheme) {
      return schemes[scheme].type;
    });

    return intersectable.some(function (item) {
      return item === preference;
    });
  });

  // Return an essentially random but consistent scheme.
  if (!method) {
    return schemes[Object.keys(schemes)[0]];
  }

  // Find the scheme that matched our preferred method.
  return schemes[Object.keys(schemes).find(function (key) {
    return schemes[key].type === method;
  })];
};

/**
 * Retrieve authentication tokens and method any way possible. It will attempt
 * to resolve automatically. If that is not possible, it will defer to
 * prompting the user.
 *
 * @param {Object}   schemes
 * @param {Object}   options
 * @param {Function} done
 */
var retrieveTokens = function retrieveTokens(schemes, options) {
  var scheme = preferredScheme(schemes);

  if (!scheme) {
    return Promise.reject(new Error('No schemes available'));
  }

  return promptTokens(scheme, options).then(function (tokens) {
    return { scheme: scheme, tokens: tokens };
  });
};

/**
 * Attempt to magically resolve to the first working authentication method. If
 * we fail, we need to fall back to manual authentication options with the
 * optimal authentication scheme available.
 *
 * @param {Object}   schemes
 * @param {Object}   options
 * @param {Function} done
 */
var resolveScheme = function resolveScheme(schemes, options) {
  return retrieveTokens(schemes, options).then(function (_ref) {
    var scheme = _ref.scheme,
        tokens = _ref.tokens;
    return authenticate(scheme, tokens);
  });
};

/**
 * Export a function that will contain all the logic for automagically
 * selecting an appropriate authentication method and prompting the user
 * for the following steps.
 *
 * @param {Object}   schemes
 * @param {String}   method
 * @param {Object}   options
 */
module.exports = function (schemes, method, options) {
  var auth = Object.assign({}, options);

  // If no authentication method has been passed in, attempt to pick our own.
  if (!method) {
    return resolveScheme(schemes, auth);
  }

  // Ensure we are attempting to authenticate with a valid method.
  if (!Object.prototype.hasOwnProperty.call(schemes, method)) {
    return Promise.reject(new Error('The only available authentication methods are: ' + Object.keys(schemes).map(JSON.stringify).join(', ')));
  }

  var scheme = schemes[method];

  // Finally we have everything we need and can initiate authentication.
  return authenticate(scheme, auth);
};

/**
 * Expose secret field names.
 */
module.exports.SECRET_FIELDS = SECRET_FIELDS;
//# sourceMappingURL=authenticate.js.map