'use strict';

var qs = require('querystring');
var url = require('url');

var authWindow = require('../authWindow');
var oAuth1Request = require('./request');
var request = require('../../ajax');

/**
 * Simple constant for the url encoded content type.
 *
 * @type {String}
 */
var URL_ENCODED = 'application/x-www-form-urlencoded';

/**
 * Send a request to get the initial OAuth request tokens.
 *
 * @param {String}   url
 * @param {Function} done
 */
var getRequestToken = function getRequestToken(options) {
  return new Promise(function (resolve, reject) {
    try {
      oAuth1Request({
        url: options.requestTokenUri,
        proxy: options.proxy,
        xhr: options.xhr,
        method: 'POST',
        oauth1: Object.assign({
          oauthCallback: options.redirectUri
        }, options),
        headers: {
          'Content-Type': URL_ENCODED
        }
      }, function (err, xhr) {
        if (err) {
          reject(err);
          return;
        }

        if (xhr.statusCode !== 200) {
          reject(new Error(xhr.body || 'Invalid Request'));
          return;
        }

        var content = qs.parse(xhr.body);

        resolve({
          oauthToken: content.oauth_token,
          oauthTokenSecret: content.oauth_token_secret
        });
      });
    } catch (err) {
      reject(err);
    }
  });
};

/**
 * Get the final access token by passing in the request options and the token
 * verifier.
 *
 * @param {Object}   options
 * @param {String}   verifier
 * @param {Function} done
 */
var getAccessToken = function getAccessToken(options, verifier) {
  return new Promise(function (resolve, reject) {
    request({
      url: options.tokenCredentialsUri,
      proxy: options.proxy,
      xhr: options.xhr,
      method: 'POST',
      oauth1: options,
      headers: {
        'Content-Type': URL_ENCODED
      },
      sync: false,
      body: qs.stringify({
        oauth_verifier: verifier
      })
    }, function (err, xhr) {
      if (err) {
        reject(err);
        return;
      }

      try {
        var response = qs.parse(xhr.body);

        var data = {
          response: response,
          oauthToken: response.oauth_token,
          oauthTokenSecret: response.oauth_token_secret
        };

        // Delete data that has been pulled off the response object to avoid
        // duplication.
        delete response.oauth_token;
        delete response.oauth_token_secret;

        if (!Object.keys(data.response).length) {
          delete data.response;
        }

        resolve(data);
      } catch (error) {
        reject(error);
      }
    });
  });
};

module.exports = oAuth1Flow;

/**
 * Trigger the full OAuth1 authentication flow.
 *
 * @param {Object}   options
 * @param {Function} done
 */
function oAuth1Flow(options) {
  if (typeof options.consumerKey !== 'string') {
    return Promise.reject(new TypeError('"consumerKey" expected'));
  }

  if (typeof options.consumerSecret !== 'string') {
    return Promise.reject(new TypeError('"consumerSecret" expected'));
  }

  if (typeof options.requestTokenUri !== 'string') {
    return Promise.reject(new TypeError('"requestTokenUri" expected'));
  }

  if (typeof options.authorizationUri !== 'string') {
    return Promise.reject(new TypeError('"authorizationUri" expected'));
  }

  if (typeof options.tokenCredentialsUri !== 'string') {
    return Promise.reject(new TypeError('"tokenCredentialsUri" expected'));
  }

  return getRequestToken(options).then(function (data) {
    return new Promise(function (resolve, reject) {
      var popup = authWindow(options.authorizationUri + '?' + qs.stringify({
        oauth_token: data.oauthToken
      }), function (err) {
        if (err) {
          reject(err);
        }

        resolve();
      });

      global.authenticateOAuth = function (href) {
        popup.close();
        delete global.authenticateOAuth;

        if (href.substr(0, options.redirectUri.length) !== options.redirectUri) {
          reject(new Error('Invalid redirect URI'));
          return;
        }

        var response = url.parse(href, true).query;

        if (response.oauth_token !== data.oauthToken) {
          reject(new Error('Invalid OAuth token response'));
          return;
        }

        getAccessToken(Object.assign({
          oauthToken: response.oauth_token
        }, options), response.oauth_verifier).then(function (accessToken) {
          return resolve(accessToken);
        });
      };
    });
  });
}
//# sourceMappingURL=index.js.map