'use strict';

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var qs = require('querystring');
var url = require('url');
var createHmac = require('create-hmac');
var request = require('../../ajax');

var _require = require('../../utils'),
    isObject = _require.isObject;

var _require2 = require('./utils'),
    encodeData = _require2.encodeData,
    normalizeUrl = _require2.normalizeUrl;

module.exports = oAuth1Request;

/**
 * Simple constant for the url encoded content type.
 *
 * @type {String}
 */
var URL_ENCODED = 'application/x-www-form-urlencoded';

/**
 * Check whether the parameters name is a valid OAuth parameter.
 *
 * @param  {String}  param
 * @return {Boolean}
 */
var isParamAnOAuthParameter = function isParamAnOAuthParameter(param) {
  return (/^oauth_/.test(param)
  );
};

/**
 * Transform an object of key, value pairs to an array of arrays.
 *
 * @param  {Object} obj
 * @return {Array}
 */
var paramsToArray = function paramsToArray(obj) {
  return Object.keys(obj).map(function (key) {
    return [key, obj[key]];
  });
};

/**
 * Generate the Authorization header from an order parameter array.
 *
 * @param  {Object} data
 * @param  {Array}  params
 * @return {String}
 */
var buildAuthorizationHeaders = function buildAuthorizationHeaders(data, params) {
  return 'OAuth realm="' + normalizeUrl(data.url) + '",' + params.filter(function (param) {
    return isParamAnOAuthParameter(param[0]);
  }).map(function (param) {
    return param[0] + '="' + param[1] + '"';
  }).join(',');
};

/**
 * Return the current timestamp in seconds since January 1, 1970 00:00:00 GMT.
 *
 * @return {Number}
 */
var getTimestamp = function getTimestamp() {
  return Math.floor(Date.now() / 1000);
};

/**
 * Generate a random nonce string.
 *
 * @return {String}
 */
var getNonce = function getNonce() {
  var chars = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9'];

  var nonce = '';

  for (var i = 0; i < 32; i += 1) {
    nonce += chars[Math.floor(Math.random() * chars.length)];
  }

  return nonce;
};

/**
 * Transform an array of parameters (in nested array form) to a query string.
 *
 * @param  {Array}  array
 * @return {String}
 */
var arrayToParams = function arrayToParams(array) {
  return array.map(function (args) {
    return args[0] + '=' + args[1];
  }).join('&');
};

/**
 * Sort query string parameters (represented as an array of arrays) by name and
 * then value.
 *
 * @param  {Array} argPairs
 * @return {Array}
 */
var sortRequestParams = function sortRequestParams(argPairs) {
  return argPairs.sort(function (a, b) {
    if (a[0] === b[0]) {
      return a[1] < b[1] ? -1 : 1;
    }

    return a[0] < b[0] ? -1 : 1;
  });
};

/**
 * Create the base signature string for hashing.
 *
 * @param  {Object} data
 * @return {String}
 */
var createSignatureBase = function createSignatureBase(params, data) {
  return [data.method.toUpperCase(), encodeData(normalizeUrl(data.url)), encodeData(arrayToParams(params))].join('&');
};

/**
 * Generate a signature string combining the base signature with consumer
 * secrets.
 *
 * @param  {String} base
 * @param  {Object} data
 * @return {String}
 */
var createSignature = function createSignature(base, options) {
  var key = [encodeData(options.consumerSecret), encodeData(options.oauthTokenSecret)].join('&');

  var hash = key;

  if (options.signatureMethod === 'HMAC-SHA1') {
    hash = createHmac('sha1', key).update(base).digest('base64');
  }

  return hash;
};

/**
 * Generate a signature from the AJAX data.
 *
 * @param  {Array}  params
 * @param  {Object} data
 * @return {String}
 */
var getSignature = function getSignature(params, data) {
  var signatureBase = createSignatureBase(params, data);
  return createSignature(signatureBase, data.oauth1);
};

/**
 * Prepare ordered OAuth parameters based on the provided data.
 *
 * @param  {Object} data
 * @return {Array}
 */
var prepareParameters = function prepareParameters(data) {
  var params = paramsToArray(Object.assign({
    oauth_timestamp: getTimestamp(),
    oauth_nonce: getNonce(),
    oauth_version: '1.0',
    oauth_signature_method: data.oauth1.signatureMethod,
    oauth_consumer_key: data.oauth1.consumerKey
  }, data.url.query));

  // Attach the token query parameter if we have one.
  if (data.oauth1.oauthToken) {
    params.push(['oauth_token', data.oauth1.oauthToken]);
  }

  if (data.oauth1.oauthCallback) {
    params.push(['oauth_callback', data.oauth1.oauthCallback]);
  }

  var contentType = Object.keys(data.headers || {}).map(function (key) {
    return [key, data.headers[key]];
  }).find(function (header) {
    return header[0].toLowerCase() === 'content-type';
  });

  if (!contentType) {
    contentType = URL_ENCODED;
    data.headers['Content-Type'] = contentType;
  } else {
    // eslint-disable-next-line prefer-destructuring
    contentType = contentType[1];
  }

  if (contentType === URL_ENCODED) {
    if (typeof data.body === 'string') {
      data.body = qs.parse(data.body);
    }

    if (isObject(data.body)) {
      var body = paramsToArray(data.body);
      data.body = arrayToParams(body);
      params.push.apply(params, _toConsumableArray(body));
    }
  }

  var sortedParams = sortRequestParams(params);

  sortedParams.push(['oauth_signature', encodeData(getSignature(sortedParams, data))]);

  return sortedParams;
};

/**
 * Allow a new ajax flow for OAuth1-based URLs. Accepts an `oauth1` property
 * on the data object in the format that is returned from the middleware.
 *
 * @param {Object}   data
 * @param {Function} next
 */

function oAuth1Request(data, done) {
  // Check we have an oauth1 object for attempting to mixin keys.
  if (isObject(data.oauth1)) {
    if (!data.oauth1.signatureMethod) {
      data.oauth1.signatureMethod = 'HMAC-SHA1';
    }

    // Parse the url for augmenting the query string parameters. Needed in
    // multiple places throughout the flow, so we can minimize the number of
    // parses by doing it once at the start.
    data.url = url.parse(data.url, true);

    // Delete parameters specific to re-adding the query string, since we need
    // to regenerate the query string without OAuth params.
    delete data.url.href;
    delete data.url.path;
    delete data.url.search;

    var orderedParams = prepareParameters(data);
    var authorization = buildAuthorizationHeaders(data, orderedParams);

    data.headers.Authorization = authorization;

    data.url.query = arrayToParams(paramsToArray(data.url.query).filter(function (param) {
      return !isParamAnOAuthParameter(param[0]);
    }));

    // Reattach the query string if we have one available.
    if (data.url.query) {
      data.url.search = '?' + data.url.query;
      data.url.path = data.url.pathname + data.url.search;
    }

    data.sync = true;
    data.url = url.format(data.url);
  }

  return request(data, done);
}
//# sourceMappingURL=request.js.map