'use strict';

var qs = require('querystring');
var template = require('./template');
var sanitizeAST = require('./sanitizeAst');
var buildDocumentation = require('./documentation');
var basicAuthRequest = require('../authentication/basic/request');
var oAuth1Request = require('../authentication/oauth1/request');
var oAuth2Request = require('../authentication/oauth2/request');
var ajaxRequest = require('../ajax');

var _require = require('../utils'),
    isObject = _require.isObject;

var CONFIG_PROPERTY = '!config';
var CLIENT_PROPERTY = '!client';
var JSON_REGEXP = /^application\/([\w!#$%&*`\-.^~]*\+)?json$/i;

var HTTP_METHODS = ['get', 'head', 'put', 'post', 'patch', 'delete'];
var CONFIG_OPTIONS = ['body', 'sync', 'proxy', 'xhr', 'uriParameters', 'baseUri', 'baseUriParameters', 'headers', 'query', 'beforeSend'];
var OVERRIDABLE_CONFIG_OPTIONS = ['body', 'sync', 'proxy', 'xhr', 'baseUri', 'beforeSend'];

/**
 * Map the supported auth types to the known triggers.
 *
 * @type {Object}
 */
var authMap = {
  'OAuth 1.0': 'oauth1',
  'OAuth 2.0': 'oauth2',
  'Basic Authentication': 'basicAuth'
};

/**
 * Transform a data object into a form data instance.
 *
 * @param  {Object}   data
 * @return {FormData}
 */
var toFormData = function toFormData(data) {
  var form = new FormData();

  // Iterate over every piece of data and append to the form data object.
  if (data) {
    Object.keys(data).forEach(function (key) {
      var value = data[key];

      form.append(key, value);
    });
  }

  return form;
};

/**
 * Transform a data object into a String.
 *
 * @param  {Object}   data
 * @return {String}
 */
var toString = function toString(data) {
  return String(data);
};

/**
 * Map mime types to their parsers.
 *
 * @type {Array}
 */
var parse = [[JSON_REGEXP, JSON.parse], ['application/x-www-form-urlencoded', qs.parse]];

/**
 * Map mime types to their serializers.
 *
 * @type {Array}
 */
var serialize = [[JSON_REGEXP, JSON.stringify], ['application/x-www-form-urlencoded', qs.stringify], ['text/html', toString], ['multipart/form-data', toFormData]];

/**
 * Iterate over an array of match and result values, and return the
 * first matching value.
 *
 * @param  {Array}    array
 * @param  {String}   test
 * @return {Function}
 */
var getMatch = function getMatch(array, test) {
  var match = array.find(function (value) {
    var check = value[0];

    if (check instanceof RegExp) {
      return check.test(test);
    }

    return check === test;
  });

  return match && match[1];
};

/**
 * Sort mimes by preference.
 *
 * @param  {String} mime
 * @return {Number}
 */
var mimePreference = function mimePreference(mime) {
  return getMatch([[JSON_REGEXP, 3], ['application/x-www-form-urlencoded', 2], ['multipart/form-data', 1]], mime) || 0;
};

/**
 * Check if a method is a query method (not a body as the argument).
 *
 * @param  {String}  method
 * @return {Boolean}
 */
function isQueryMethod(method) {
  return method === 'get' || method === 'head';
}

/**
 * List of all plain HTTP methods in the format from the AST.
 *
 * @type {Object}
 */
var allHttpMethods = HTTP_METHODS.reduce(function (acum, current) {
  acum[current] = {
    method: current
  };

  return acum;
}, {});

/**
 * Return the xhr response mime type.
 *
 * @param  {String} contentType
 * @return {String}
 */
var getMime = function getMime(contentType) {
  return contentType == null ? null : String(contentType).split(';')[0];
};

/**
 * Check if an object is a host object and avoid serializing.
 *
 * @param  {Object}  obj
 * @return {Boolean}
 */
var isHost = function isHost(obj) {
  var str = Object.prototype.toString.call(obj);

  switch (str) {
    case '[object File]':
    case '[object Blob]':
    case '[object String]':
    case '[object Number]':
    case '[object Boolean]':
    case '[object FormData]':
      return true;
    default:
      return false;
  }
};

/**
 * Sanitize a specific configuration option.
 *
 * @type {Object}
 */
var sanitizeOption = {
  query: function query(_query) {
    if (typeof _query === 'string') {
      return qs.parse(_query);
    }

    return _query;
  }
};

/**
 * Gets a header from the header object.
 *
 * @param  {Object}  headers
 * @param  {String}  header
 * @return {String}
 */
var findHeader = function findHeader(headers, header) {
  if (!headers) {
    return null;
  }

  header = header.toLowerCase();

  var key = Object.keys(headers).find(function (name) {
    return name.toLowerCase() === header;
  });

  return headers[key];
};

/**
 * Sanitize the XHR request into the desired format.
 *
 * @param  {XMLHttpRequest} xhr
 * @return {Object}
 */
var sanitizeXHR = function sanitizeXHR(xhr) {
  if (!xhr) {
    return xhr;
  }

  var mime = getMime(xhr.headers['content-type']);
  var body = xhr.body.trim();

  // Automatically parse all response bodies. Only parse when we have response
  // body content. Errors that occur parsing will be propagated to the client.
  body = body ? (getMatch(parse, mime) || function (value) {
    return value;
  })(body) : undefined;

  return {
    body: body,
    status: xhr.statusCode,
    headers: xhr.headers
  };
};

/**
 * Returns a function that can be used to make ajax requests.
 *
 * @param  {Object}   nodes
 * @param  {Object}   method
 * @return {Function}
 */
var httpRequest = function httpRequest(nodes, method) {
  return function (body, config, async) {
    config = config || {};

    // Map configuration options and merge with the passed in object.
    config = CONFIG_OPTIONS.reduce(function (acum, option) {
      if (OVERRIDABLE_CONFIG_OPTIONS.indexOf(option) >= 0) {
        acum[option] = config[option] ? config[option] : nodes.config[option];

        return acum;
      }

      var nodeOption = nodes.config[option];
      var configOption = config && config[option];
      var sanitize = sanitizeOption[option] || function (value) {
        return value;
      };

      acum[option] = Object.assign({}, sanitize(nodeOption), sanitize(configOption));

      return acum;
    }, {});

    var request = ajaxRequest;
    var mime = getMime(findHeader(config.headers, 'Content-Type'));
    var baseUri = template(config.baseUri, config.baseUriParameters);
    var fullUri = baseUri.replace(/\/+$/, '') + '/' + nodes.join('/');

    // GET and HEAD requests accept the query string as the first argument.
    if (isQueryMethod(method.method)) {
      Object.assign(config.query, sanitizeOption.query(body));
      body = null;
    }

    // Set the config object body to the passed in body.
    if (body != null) {
      config.body = body;
    }

    // Append the query string if one is available.
    if (Object.keys(config.query).length) {
      fullUri += '?' + qs.stringify(config.query);
    }

    // If we have no accept header set already, default to accepting
    // everything. This is required because Firefox sets the base accept
    // header to essentially be `html/xml`.
    if (!findHeader(config.headers, 'accept')) {
      config.headers.accept = '*/*';
    }

    // If we were passed in data, attempt to sanitize it to the correct type.
    if (!isHost(config.body)) {
      // Set the correct `Content-Type` header, if none exists. Kind of random
      // if more than one exists - I would suggest setting it yourself.
      if (mime == null) {
        // If we have a method body object, sort the method types by most
        // desirable and fallback to a random content type.
        if (isObject(method.body)) {
          mime = Object.keys(method.body).sort(mimePreference).pop();
        }

        // Set the config to the updated mime type header. If none exists, use
        // `application/json` by default.
        mime = mime || 'application/json';
        config.headers['Content-Type'] = mime;
      }

      // Automatically attempt to serialize the body.
      var serializer = getMatch(serialize, mime);

      if (!serializer) {
        throw new TypeError('Can not serialize content type of "' + mime + '"');
      }

      try {
        config.body = serializer(config.body);
      } catch (e) {
        throw new TypeError('Could not serialize body: ' + e.message);
      }
    }

    if (async !== undefined) {
      config.sync = !async;
    }

    var options = {
      url: fullUri,
      body: config.body,
      sync: config.sync,
      xhr: config.xhr,
      proxy: config.proxy,
      method: method.method,
      headers: config.headers,
      beforeSend: config.beforeSend
    };

    // Iterate through `securedBy` methods and accept the first one we are
    // already authenticated for.
    var securedBy = Object.keys(method.securedBy || nodes.client.securedBy || {});

    for (var i = 0; i < securedBy.length; i += 1) {
      var key = securedBy[i];
      var scheme = nodes.client.securitySchemes[key];

      // Scheme is not documented in the RAML security schemes.
      if (!scheme) {
        continue;
      }

      var authenticated = nodes.client.authentication[scheme.type];
      var authType = authMap[scheme.type];

      if (authenticated) {
        options[authType] = authenticated;

        if (authType === 'oauth1') {
          request = oAuth1Request;
        }

        if (authType === 'oauth2') {
          request = oAuth2Request;
        }

        if (authType === 'basicAuth') {
          request = basicAuthRequest;
        }

        break;
      }

      continue;
    }

    if (config.sync) {
      // Awkward sync and async code mixing.
      var response = void 0;
      var error = void 0;

      request(options, function (err, xhr) {
        if (err) {
          error = err;
          return;
        }

        try {
          response = sanitizeXHR(xhr);
        } catch (innerError) {
          error = innerError;
        }
      });

      if (error) {
        throw error;
      }

      return response;
    }

    return new Promise(function (resolve, reject) {
      try {
        request(options, function (err, xhr) {
          if (err) {
            reject(err);
            return;
          }

          resolve(xhr);
        });
      } catch (err) {
        reject(err);
      }
    }).then(sanitizeXHR);
  };
};

/**
 * Attaches XHR request methods to the context object for each available method.
 *
 * @param  {Array}  nodes
 * @param  {Object} context
 * @param  {Object} methods
 * @return {Object}
 */
var attachMethods = function attachMethods(nodes, context, methods) {
  if (!methods) {
    return context;
  }

  // Attach the available methods to the current context.
  Object.keys(methods).forEach(function (verb) {
    var method = methods[verb];

    context[verb] = httpRequest(nodes, method);
  });

  return context;
};

/**
 * Attach a special media extension handler.
 *
 * @param  {Array}  nodes
 * @param  {Object} context
 * @param  {Object} resource
 * @return {Object}
 */
var attachMediaTypeExtension = function attachMediaTypeExtension(nodes, context, resource) {
  /**
   * Push the extension onto the current route and set relevant headers.
   *
   * @param  {String} extension
   * @return {Object}
   */
  context.extension = function (extension) {
    extension = extension == null ? '' : String(extension);

    // Prepend a period to the extension before adding to the route.
    if (extension && extension.charAt(0) !== '.') {
      extension = '.' + extension;
    }

    var newContext = {};
    var routeNodes = Object.assign([], nodes);

    // Append the extension to the current route.
    routeNodes[routeNodes.length - 1] += extension;

    attachMethods(routeNodes, newContext, resource.methods);
    attachResources(routeNodes, newContext, resource.resources);

    return newContext;
  };

  // If the media type extension is not required, allow direct invocation of
  // the request method.
  if (resource.uriParameters.mediaTypeExtension.required === false) {
    attachMethods(nodes, context, resource.methods);
  }

  // Iterate over the enum options and automatically attach to the context.
  if (resource.uriParameters.mediaTypeExtension.enum) {
    Object.keys(resource.uriParameters.mediaTypeExtension.enum).forEach(function (key) {
      var extension = resource.uriParameters.mediaTypeExtension.enum[key];

      if (extension.charAt(0) === '.') {
        extension = extension.substr(1);
      }

      context[extension] = context.extension(extension);
    });
  }

  return context;
};

/**
 * Generate a context or attach methods and resources to an existing context.
 *
 * @param  {Array}   nodes
 * @param  {Object}  resource
 * @param  {Boolean} hasMediaExtension
 * @param  {Object}  [context]
 * @return {Object}
 */
var newContext = function newContext(nodes, resource, hasMediaExtension, context) {
  context = context || {};

  if (hasMediaExtension) {
    attachMediaTypeExtension(nodes, context, resource);
  } else {
    attachMethods(nodes, context, resource.methods);
    attachResources(nodes, context, resource.resources);
  }

  return context;
};

/**
 * Recurses through a resource object in the RAML AST, generating a dynamic
 * DSL that only allows methods that were defined in the RAML spec.
 *
 * @param  {Array}  nodes
 * @param  {Object} context
 * @param  {Object} resources
 * @return {Object}
 */
function attachResources(nodes, context, resources) {
  if (!resources) {
    return context;
  }

  Object.keys(resources).forEach(function (route) {
    var resource = resources[route];
    var routeNodes = Object.assign([], nodes);
    var routeName = route;
    var hasMediaExtension = route.substr(-20) === '{mediaTypeExtension}';

    // Ignore media type extensions in route generation.
    if (hasMediaExtension) {
      routeName = route.slice(0, -20);
      route = routeName;
    }

    // Check the route against our valid uri parameters.
    var templateTags = route.match(template.REGEXP);

    // Push the current route into the route array.
    routeNodes.push(route);

    // If we have template tags available, attach a dynamic route.
    if (templateTags) {
      var routeSuffix = templateTags.join('');

      // The route must end with the chained template tags and have no
      // text between tags.
      if (route.substr(-routeSuffix.length) !== routeSuffix) {
        return;
      }

      // If the route is only a template tag with no static text, use the
      // template tag text as the method name.
      if (templateTags.length === 1 && route === templateTags[0]) {
        routeName = templateTags[0].slice(1, -1);
      } else {
        routeName = route.substr(0, route.indexOf('{'));
      }

      // Avoid adding empty route name cases. This can occur when we have
      // multiple tag names and no front text. For example, `{this}{that}`.
      // This could also occur if for some reason we are passing in a route that
      // isn't dynamic.
      if (!routeName) {
        return;
      }

      // Get the ordered tag names for completion.
      var tags = templateTags.map(function (param) {
        return resource.uriParameters[param.slice(1, -1)];
      });

      // The route is dynamic, so we set the route name to be a function
      // which accepts the template arguments and updates the path fragment.
      // We'll extend any route already at the same namespace so we can do
      // things like use both `/{route}` and `/route`, if needed.
      context[routeName] = Object.assign(function () {
        for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
          args[_key] = arguments[_key];
        }

        // Map the tags to the arguments or default arguments.
        var parts = tags.map(function (tag, index) {
          // Inject enum parameters if there is only one available enum.
          // TODO: When/if we add validation back, have these routes
          // be generated instead of typed out.
          if (args[index] == null && tag.enum && tag.enum.length === 1) {
            return tag.enum[0];
          }

          // Use any passed in argument - even it's falsy.
          if (index in args) {
            return args[index];
          }

          var param = templateTags[index].slice(1, -1);

          // Fallback to injecting the fallback configuration uri parameter.
          return routeNodes.config && routeNodes.config.uriParameters[param];
        });

        // Change the last path fragment to the proper template text.
        routeNodes[routeNodes.length - 1] = template(route, parts);

        return newContext(routeNodes, resource, hasMediaExtension);
      }, context[routeName]);
    }

    // Handle root `/` resources.
    if (!routeName) {
      context = newContext(routeNodes, resource, hasMediaExtension, context);
    } else {
      context[routeName] = newContext(routeNodes, resource, hasMediaExtension, context[routeName]);
    }
  });

  return context;
}

/**
 * Generate the client object from a sanitized AST object.
 *
 * @param  {Object} ast Passed through `sanitizeAST`
 * @param  {Object} config
 * @return {Object}
 */
var generateClient = function generateClient(name, ast, config) {
  var editor = config.editor;

  // Generate the base uri parameters and defaults.

  var baseUriParameters = Object.assign({ version: ast.version }, config.baseUriParameters);

  // Generate the root node array. Set properties directly on this array to be
  // copied to the next execution part. We have a global configuration object
  // which can be altered externally at any point, as well as when we finally
  // make a request. For this reason, it's important that we use objects which
  // are passed by reference.
  var nodes = Object.assign([], {
    config: Object.assign({
      baseUri: ast.baseUri
    }, config, {
      baseUriParameters: baseUriParameters
    }),
    client: {
      securedBy: ast.securedBy,
      authentication: {},
      securitySchemes: ast.securitySchemes,
      baseUriParameters: ast.baseUriParameters
    }
  });

  // Throw an error if the baseUri string is missing.
  if (typeof nodes.config.baseUri !== 'string') {
    throw new Error('A "baseUri" string is required');
  }

  /**
   * The root client implementation is simply a function. This allows us to
   * enter a custom path that may not be supported by the DSL and run any
   * method regardless of whether it was defined in the spec.
   *
   * @param  {String} path
   * @param  {Object} context
   * @return {Object}
   */
  var client = function client(path, context) {
    var route = template(path || '', context || {}).replace(/^\/+/, '').split('/');

    return attachMethods(Object.assign([], nodes, route), {}, allHttpMethods);
  };

  client[CONFIG_PROPERTY] = nodes.config;
  client[CLIENT_PROPERTY] = nodes.client;
  client.toDescription = function () {
    return 'Client has been successfully created';
  };

  attachResources(nodes, client, ast.resources);

  if (editor && !editor.languages.typescript.javascriptDefaults.getExtraLibs()[name + '-client.d.ts']) {
    editor.languages.typescript.javascriptDefaults.addExtraLib(buildDocumentation(name, ast), name + '-client.d.ts');
  }

  return client;
};

/**
 * Exports the client generator, which accepts the AST of a RAML document.
 *
 * @return {Object} Dynamic object for constructing API requests from the AST.
 */
module.exports = function (name, ast, config) {
  return generateClient(name, sanitizeAST(ast), config);
};
//# sourceMappingURL=index.js.map