'use strict';

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

/**
 * Sanitize the secured by into an object.
 *
 * @param  {Array}  secured
 * @return {Object}
 */
var sanitizeSecuredBy = function sanitizeSecuredBy(secured) {
  if (!Array.isArray(secured)) {
    return null;
  }

  var securedBy = {};

  // Since `securedBy` can either be an array of strings or array of objects
  // with only a single key, we merge it down to an object.
  secured.forEach(function (value) {
    if (value == null) {
      return; // Ignore `null` array values, shouldn't be useful to me.
    }

    if (typeof value === 'string') {
      securedBy[value] = true;
      return;
    }

    Object.assign(securedBy, value);
  });

  return securedBy;
};

/**
 * Sanitize the AST from the RAML parser into something easier to work with.
 *
 * @param  {Object} ast
 * @return {Object}
 */
module.exports = function (ast) {
  // Create the base sanitized ast with only the properties we want.
  var sanitizedAst = {
    title: ast.title,
    version: ast.version,
    baseUri: ast.baseUri,
    baseUriParameters: ast.baseUriParameters
  };

  // Merge an array of objects into a single object using `Object.assign` and
  // `apply` (since `Object.assign` accepts unlimited number of arguments).
  if (ast.securitySchemes) {
    sanitizedAst.securitySchemes = Object.assign.apply(Object, _toConsumableArray(ast.securitySchemes));
  }

  // Sanitize secured by which is a bit more complicated than extending.
  if (ast.securedBy) {
    sanitizedAst.securedBy = sanitizeSecuredBy(ast.securedBy);
  }

  // Recurse through the resources and move URIs to be the key names.
  sanitizedAst.resources = function flattenResources(resources) {
    var map = {};

    // Resources are provided as an array, we'll move them to be an object.
    resources.forEach(function (resource) {
      var sanitizedResource = {};

      // Methods are implemented as arrays of objects too, but not recursively.
      // TODO: If the endpoint is the final route and has no methods, implement
      // backtracking and remove access to it from the AST.
      if (resource.methods) {
        sanitizedResource.methods = resource.methods.reduce(function (acum, current) {
          // Create the sanitized method by including the properties we want.
          acum[current.method] = {
            method: current.method,
            body: current.body,
            headers: current.headers,
            description: current.description,
            queryParameters: current.queryParameters,
            queryString: current.queryString
          };

          // Sanitize the `securedBy` method.
          if (current.securedBy) {
            acum[current.method].securedBy = sanitizeSecuredBy(current.securedBy);
          }

          return acum;
        }, {});
      }

      if (resource.resources) {
        sanitizedResource.resources = flattenResources(resource.resources);
      }

      (function attachResource($map, segments) {
        var segment = segments.shift();
        var part = void 0;

        $map[segment] = $map[segment] || {};
        part = $map[segment];

        // Currently on the last url segment, embed the full resource.
        if (!segments.length) {
          $map[segment] = sanitizedResource;
          part = $map[segment];
        }

        // Pull any possible tags out of the relative uri part.
        var match = segment.match(/\{([^}]+)\}/g);
        var tags = match ? match.map(function (tag) {
          return tag.slice(1, -1);
        }) : [];

        // Add only the used tags to the current resource segment.
        if (tags.length) {
          part.uriParameters = tags.reduce(function (acum, current) {
            acum[current] = resource.uriParameters[current];

            return acum;
          }, {});
        }

        // If we have more segment parts left, recursively embed resources.
        if (segments.length) {
          part.resources = part.resources || {};
          attachResource(part.resources, segments);
        }
      })(map, resource.relativeUri.substr(1).split('/'));
    });

    return map;
  }(ast.resources);

  return sanitizedAst;
};
//# sourceMappingURL=sanitizeAst.js.map