const qs = require('querystring');
const url = require('url');
const { authResponse, sanitizeOptions } = require('./utils');
const authWindow = require('../authWindow');

module.exports = oAuth2TokenFlow;

/**
 * Trigger the client-side implicit OAuth2 flow.
 * Reference: http://tools.ietf.org/html/rfc6749#section-4.2
 *
 * @param {Object}   options
 * @param {Function} done
 */
function oAuth2TokenFlow(options) {
  options = sanitizeOptions(options);

  if (typeof options.clientId !== 'string') {
    return Promise.reject(new TypeError('"clientId" expected'));
  }

  if (typeof options.authorizationUri !== 'string') {
    return Promise.reject(new TypeError('"authorizationUri" expected'));
  }

  return new Promise((resolve, reject) => {
    const state = (`${Math.random()}`).substr(2);
    const popup = authWindow(`${options.authorizationUri}?${qs.stringify({
      state,
      scope: options.scope,
      client_id: options.clientId,
      response_type: 'token',
      redirect_uri: options.redirectUri
    })}`, (err) => {
      if (err) {
        reject(err);
      }

      resolve();
    });

    global.authenticateOAuth = function (href) {
      popup.close();
      delete global.authenticateOAuth;

      const uri = url.parse(href, true);
      const response = Object.assign(qs.parse((uri.hash || '').substr(1)), uri.query);

      if (href.substr(0, options.redirectUri.length) !== options.redirectUri) {
        reject(new Error('Invalid redirect uri'));
        return;
      }

      if (response.state !== state) {
        reject(new Error('State mismatch'));
        return;
      }

      // Pass the response off for validation. At least Instagram has a bug where
      // the state is being passed back as part of the query string instead of the
      // hash, so we merge both options together.
      resolve(authResponse(options, response));
    };
  });
}
